# SPDX-License-Identifier: MIT
# Copyright (c) 2023 Imagination Technologies Ltd. All Rights Reserved

'''
Test error cases.
'''

from random import Random

from constrainedrandom import RandObj, RandomizationError

from .basic import MultiSum
from .. import testutils


class ImpossibleThorough(MultiSum):
    '''
    Test the thorough solver for a problem that will always fail.

    The thorough solver almost always converges, so it's very
    hard to construct a problem 'too hard' for it.
    Instead, make it impossible and only enable the thorough
    solver.
    '''

    EXPECT_FAILURE = True

    def get_randobj(self, *args):
        randobj = super().get_randobj(*args)
        # Only use thorough solver.
        randobj.set_solver_mode(naive=False, sparse=False)
        # Make problem impossible so that numbers can't sum to 41.
        def mod_2(z):
            return z % 2 == 0
        randobj.add_constraint(mod_2, ('x',))
        randobj.add_constraint(mod_2, ('y',))
        randobj.add_constraint(mod_2, ('z',))
        # Setting max domain size to 1 makes this run a bit quicker.
        randobj._max_domain_size = 1
        return randobj


class ImpossibleOneVar(testutils.RandObjTestBase):
    '''
    Test an impossible constraint problem with one variable.
    '''

    def get_randobj(self, *args):
        randobj = RandObj(*args)
        def eq_zero(x):
            return x == 0
        randobj.add_rand_var('a', domain=[1,], constraints=[eq_zero,])
        return randobj

    def test_randobj(self):
        # Override the whole test function, because unsolvable
        # simple single variables fail on creation, rather
        # than when randomize is called.
        self.assertRaises(RandomizationError, self.get_randobj, 0)
        self.assertRaises(RandomizationError, self.get_randobj, 1)


class ImpossibleComplexVar(testutils.RandObjTestBase):
    '''
    Test an impossible constraint problem with one variable, where
    the variable state space is too large to fail on creation.
    '''

    EXPECT_FAILURE = True

    def get_randobj(self, *args):
        randobj = RandObj(*args)
        def eq_minus_one(x):
            return x == -1
        randobj.add_rand_var('a', bits=64, constraints=[eq_minus_one,])
        return randobj


class ImpossibleMultiVar(testutils.RandObjTestBase):
    '''
    Test an impossible constraint problem with multiple variables.
    '''

    EXPECT_FAILURE = True

    def get_randobj(self, *args):
        randobj = RandObj(*args)
        def lt_5(x):
            return x < 5
        randobj.add_rand_var('a', domain=range(10), constraints=[lt_5,])
        randobj.add_rand_var('b', domain=range(10), constraints=[lt_5,])
        def sum_gt_10(x, y):
            return x + y > 10
        randobj.add_constraint(sum_gt_10, ('a', 'b'))
        return randobj
