# SPDX-License-Identifier: MIT
# Copyright (c) 2023 Imagination Technologies Ltd. All Rights Reserved

'''
Test random lists.
'''

from random import Random

from constrainedrandom import RandObj
from constrainedrandom.utils import unique
from .. import testutils


def plus_or_minus_one(listvar):
    val = listvar[0]
    for nxt_val in listvar[1:]:
        if nxt_val == val + 1 or nxt_val == val - 1:
            val = nxt_val
            continue
        return False
    return True


def sum_0(listvar):
    return sum(listvar) == 0


class RandList(testutils.RandObjTestBase):
    '''
    Test a randomized list.
    '''

    ITERATIONS = 1000
    LENGTH = 10

    def get_randobj(self, *args):
        r = RandObj(*args)
        def not_7(x):
            return x != 7
        r.add_rand_var('listvar', domain=range(10), constraints=[not_7], length=self.LENGTH)
        return r

    def check(self, results):
        nonzero_seen = False
        for result in results:
            self.assertIsInstance(result['listvar'], list, "Var with length > 0 wasn't a list")
            self.assertEqual(len(result['listvar']), self.LENGTH, "Length incorrect")
            for x in result['listvar']:
                self.assertIn(x, range(10), "Value was wrongly randomized")
                self.assertNotEqual(x, 7, "Scalar constraint not respected")
                if x != 0:
                    nonzero_seen = True
        self.assertTrue(nonzero_seen, "All values were zero")


class RandListConstrained(testutils.RandObjTestBase):
    '''
    Test a randomized list with a basic list constraint.
    Keep it simple enough that we use the CSP list solver.
    '''

    ITERATIONS = 1000
    LENGTH = 2

    def get_randobj(self, *args):
        r = RandObj(*args)
        def not_7(x):
            return x != 7
        r.add_rand_var('listvar', domain=range(10), constraints=[not_7], length=self.LENGTH)
        def sum_lt_val(listvar):
            return sum(listvar) < (6 * self.LENGTH)
        r.add_constraint(sum_lt_val, ('listvar',))
        return r

    def check(self, results):
        nonzero_seen = False
        for result in results:
            self.assertIsInstance(result['listvar'], list, "Var with length > 0 wasn't a list")
            self.assertEqual(len(result['listvar']), self.LENGTH, "Length incorrect")
            self.assertLess(sum(result['listvar']), (8 * self.LENGTH), "List constraint not followed")
            for x in result['listvar']:
                self.assertIn(x, range(10), "Value was wrongly randomized")
                self.assertNotEqual(x, 7, "Scalar constraint not respected")
                if x != 0:
                    nonzero_seen = True
        self.assertTrue(nonzero_seen, "All values were zero")


class RandListConstrainedHard(RandListConstrained):
    '''
    Test a randomized list with a basic list constraint.
    Make it sufficiently complex that it requires the random solver,
    do this by increasing the length.
    '''

    ITERATIONS = 1000
    LENGTH = 10


class RandListConstrainedVeryHard(testutils.RandObjTestBase):
    '''
    Test a randomized list with a difficult constraint on the values in the list.
    '''

    ITERATIONS = 10
    LENGTH = 10

    def get_randobj(self, *args):
        r = RandObj(*args)
        r.add_rand_var('listvar', domain=range(10), length=self.LENGTH, list_constraints=(plus_or_minus_one,))
        return r

    def check(self, results):
        for result in results:
            self.assertIsInstance(result['listvar'], list, "Var with length > 0 wasn't a list")
            self.assertEqual(len(result['listvar']), self.LENGTH, "Length incorrect")
            prev = None
            for x in result['listvar']:
                self.assertIn(x, range(10), "Value was wrongly randomized")
                if prev is not None:
                    self.assertTrue(x == prev + 1 or x == prev - 1, "list constraint not followed")
                prev = x


class RandListMultivar(testutils.RandObjTestBase):
    '''
    Test a randomized list with constraints on the values in the list,
    also with constraints over multiple variables.
    '''

    ITERATIONS = 10
    LENGTH = 10

    def get_randobj(self, *args):
        r = RandObj(*args)
        r.add_rand_var('listvar', domain=range(10), length=self.LENGTH, list_constraints=(plus_or_minus_one,))
        # Give x slightly larger range so it is sometimes impossible
        r.add_rand_var('x', domain=range(11), order=1)
        def in_list(x, listvar):
            return x in listvar
        r.add_constraint(in_list, ('x', 'listvar'))
        return r

    def check(self, results):
        for result in results:
            self.assertIsInstance(result['listvar'], list, "Var with length > 0 wasn't a list")
            self.assertEqual(len(result['listvar']), self.LENGTH, "Length incorrect")
            prev = None
            for l in result['listvar']:
                self.assertIn(l, range(10), "Value was wrongly randomized")
                if prev is not None:
                    self.assertTrue(l == prev + 1 or l == prev - 1, "list constraint not followed")
                prev = l
            self.assertIn(result['x'], result['listvar'])


class RandListUnique(testutils.RandObjTestBase):
    '''
    Test that the unique constraint works on a random list.
    '''

    ITERATIONS = 100
    LENGTH = 10

    def get_randobj(self, *args):
        r = RandObj(*args)
        r.add_rand_var(
            'listvar',
            domain=range(10),
            length=self.LENGTH,
            list_constraints=(unique,),
            disable_naive_list_solver=True,
        )
        return r

    def check(self, results):
        for result in results:
            self.assertIsInstance(result['listvar'], list, "Var with length > 0 wasn't a list")
            self.assertEqual(len(result['listvar']), self.LENGTH, "Length incorrect")
            for x_idx, x in enumerate(result['listvar']):
                self.assertIn(x, range(10), "Value was wrongly randomized")
                for y_idx, y in enumerate(result['listvar']):
                    if x_idx != y_idx:
                        self.assertNotEqual(x, y, "List elements were not unique")


class RandListTempConstraints(testutils.RandObjTestBase):
    '''
    Test a randomized list with temporary constraints.
    '''

    ITERATIONS = 100
    LENGTH = 10

    def get_randobj(self, *args):
        r = RandObj(*args)
        def not_4(x):
            return x != 4
        def not_too_many_zeros(listvar):
            zeros = [x for x in listvar if x == 0]
            return len(zeros) < 5
        r.add_rand_var('listvar', domain=range(10), constraints=[not_4],
            length=self.LENGTH, list_constraints=[not_too_many_zeros])
        return r

    def check(self, results):
        for result in results:
            self.assertIsInstance(result['listvar'], list, "Var with length > 0 wasn't a list")
            self.assertEqual(len(result['listvar']), self.LENGTH, "Length incorrect")
            zeros = 0
            for x in result['listvar']:
                self.assertIn(x, range(10), "Value was wrongly randomized")
                self.assertNotEqual(x, 4, "Scalar constraint not respected")
                if x == 0:
                    zeros += 1
            self.assertLess(zeros, 5, "List constraint not respected")

    def get_tmp_constraints(self):
        def temp_constr(listvar):
            for x in listvar:
                if x == 5:
                    return False
            return True
        return [(temp_constr, ('listvar',))]

    def tmp_check(self, results):
        self.check(results)
        for result in results:
            for x in result['listvar']:
                self.assertNotEqual(x, 5, "Temp constraint not respected")


class RandListSumZero(testutils.RandObjTestBase):

    ITERATIONS = 100

    def get_randobj(self, *args):
        randobj = RandObj(*args)
        randobj.add_rand_var('listvar', length=10, domain=range(-10, 11))
        randobj.add_constraint(sum_0, ('listvar',))
        return randobj

    def check(self, results):
        nonzero_seen = False
        for result in results:
            sum = 0
            listvar = result['listvar']
            self.assertIsInstance(listvar, list, "Returned non-list")
            self.assertEqual(len(listvar), 10, "Length incorrect")
            for val in listvar:
                if val != 0:
                    nonzero_seen = True
                sum += val
            self.assertEqual(sum, 0, "Sum must be zero but wasn't.")
        self.assertTrue(nonzero_seen, "Should not always solve this problem with a list full of zeroes.")


class RandListSparse(testutils.RandObjTestBase):
    '''
    Test a randomized list with the sparse solver.
    Make it hard enough to fail with a depth-first search.
    '''

    ITERATIONS = 1

    def get_randobj(self, *args):
        randobj = RandObj(*args)
        randobj.set_solver_mode(naive=False, sparse=True, thorough=False)
        randobj.add_rand_var('listvar1', length=10, domain=range(-10, 11))
        randobj.add_constraint(sum_0, ('listvar1',))
        randobj.add_rand_var('listvar2', length=3, domain=range(-10, 11), list_constraints=[unique])
        def not_in_list(listvar1, listvar2):
            for x in listvar2:
                if x in listvar1:
                    return False
            return True
        randobj.add_constraint(not_in_list, ('listvar1', 'listvar2'))
        randobj.add_rand_var('listvar3', length=2, domain=range(-10,11))
        def awful_constraint(listvar1, listvar2, listvar3):
            total = 1
            for val in listvar1:
                if val != 0:
                    total = total * val
            for val in listvar2:
                if val != 0:
                    total = total * val
            for val in listvar3:
                if val != 0:
                    total = total * val
            return total % 3 == 1
        randobj.add_constraint(awful_constraint, ('listvar1', 'listvar2', 'listvar3'))

        return randobj

    def check(self, results):
        nonzero_seen = False
        for result in results:
            sum = 0
            product = 1
            listvar1 = result['listvar1']
            listvar2 = result['listvar2']
            listvar3 = result['listvar3']
            self.assertIsInstance(listvar1, list, "Returned non-list")
            self.assertEqual(len(listvar1), 10, "Length incorrect")
            self.assertIsInstance(listvar2, list, "Returned non-list")
            self.assertEqual(len(listvar2), 3, "Length incorrect")
            self.assertIsInstance(listvar3, list, "Returned non-list")
            self.assertEqual(len(listvar3), 2, "Length incorrect")
            for val in listvar1:
                if val != 0:
                    nonzero_seen = True
                    product = product * val
                sum += val
                self.assertNotIn(val, listvar2, "Lists must be disjoint")
            for val in listvar2:
                if val != 0:
                    product = product * val
            for val in listvar3:
                if val != 0:
                    product = product * val
            self.assertEqual(sum, 0, "Sum must be zero but wasn't.")
            self.assertEqual(product % 3, 1, "Product mod 3 wasn't 1.")
        self.assertTrue(nonzero_seen, "Should not always solve this problem with a list full of zeroes.")
